package be.iminds.ilabt.jfed.util;

import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.*;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.Charset;
import java.nio.file.*;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.PosixFilePermission;
import java.util.HashSet;
import java.util.Scanner;
import java.util.Set;

/**
 * IOUtils
 */
public class IOUtils {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(IOUtils.class);
    private static final int READ_TIMEOUT = 60 * 1000;

    @Nonnull
    public static String askCommandLineInput(@Nonnull String question) {
        return askCommandLineInput(question, System.in);
    }

    @Nonnull
    public static String askCommandLineInput(@Nonnull String question, @Nonnull InputStream in) {
        Scanner input = new Scanner(in);
        System.out.print(question + ": ");
        return input.nextLine();
    }

    @Nonnull
    public static char[] askCommandLinePassword(@Nonnull String question) {
        return askCommandLinePassword(question, System.in);
    }

    @Nonnull
    public static char[] askCommandLinePassword(@Nonnull String question, @Nonnull InputStream in) {
        Console cons = System.console();
        if ((cons = System.console()) != null) {
            char[] passwd = cons.readPassword("%s: ", question);
            return passwd;
        } else {
            //fallback if not a console
            Scanner input = new Scanner(in);
            System.out.print("*** everything you type might be shown *** " + question + ": ");
            return input.nextLine().toCharArray();
        }
    }

    @Nonnull
    public static String fileToString(@Nonnull String filename) throws IOException {
        return fileToString(new File(filename));
    }

    //this version is EVIL. The changes in the file caused the signature check on a credential to fail!
//    public static String fileToString(File file) throws IOException {
//        StringBuilder res = new StringBuilder();
//        BufferedReader br = null;
//        try {
//            br = new BufferedReader(new FileReader(file));
//            String line = br.readLine();
//            while (line != null) {
//                res.append(line);
//                res.append('\n');
//                line = br.readLine();
//            }
////        } catch (Exception e) {
////            throw e;
////        } catch (IOException e) {
////            throw e;
//        } finally {
//            try {
//                if (br != null) br.close();
//            } catch(IOException ignored) { }
//        }
//        return res.toString();
//    }
    @Nonnull
    public static String fileToString(@Nonnull File file) throws IOException {
        StringBuilder res = new StringBuilder();
        try (FileReader fr = new FileReader(file)) {
            char[] buffer = new char[4096];

            int read = fr.read(buffer);
            while (read != -1) {
                res.append(buffer, 0, read);
                read = fr.read(buffer);
            }
        }
        return res.toString();
    }

    @Nonnull
    public static String resourceToString(@Nonnull String resourceName, @Nonnull Class classInResourceRoot) throws IOException {
        assert resourceName != null;
        assert classInResourceRoot != null;

        StringBuilder res = new StringBuilder();
        InputStream resourceInputStream = classInResourceRoot.getResourceAsStream(resourceName);
        if (resourceInputStream == null) {
            throw new IOException("No resource named \"" + resourceName + "\" was found");
        }
        try (BufferedReader br = new BufferedReader(new InputStreamReader(resourceInputStream))) {
            String line = br.readLine();
            while (line != null) {
                res.append(line);
                res.append('\n');
                line = br.readLine();
            }
        }
        return res.toString();
    }

    @Nonnull
    public static byte[] fileToByteArray(@Nonnull File file) throws IOException {
        InputStream is = new FileInputStream(file);
        return streamToByteArray(is);
    }

    @Nonnull
    public static byte[] streamToByteArray(@Nonnull InputStream is) throws IOException {
        try (ByteArrayOutputStream baos = new ByteArrayOutputStream()) {
            int i;
            byte[] b = new byte[1024];
            while ((i = is.read(b)) != -1) baos.write(b, 0, i);
            return baos.toByteArray();
        }
    }

    public static void byteArrayToFile(@Nonnull File file, @Nonnull byte[] b) throws IOException {
        try (BufferedOutputStream os = new BufferedOutputStream(new FileOutputStream(file))) {
            os.write(b);
        }
    }

    @Nonnull
    public static String streamToString(@Nonnull final InputStream is, @Nonnull String encoding/*example "UTF-8"*/) throws IOException {
        assert is != null;

        int BUFFER_SIZE = 2048;
        final char[] buffer = new char[BUFFER_SIZE];
        final StringBuilder out = new StringBuilder();
        try {
            try (Reader in = new InputStreamReader(is, encoding)) {
                for (; ; ) {
                    int rsz = in.read(buffer, 0, buffer.length);
                    if (rsz < 0)
                        break;
                    out.append(buffer, 0, rsz);
                }
            }
        } catch (UnsupportedEncodingException e) {
            throw new RuntimeException("Unexpected unsupported encoding \""+encoding+"\"", e);
        }
        return out.toString();
    }

    @Nonnull
    public static String urlToString(@Nonnull URL url, @Nonnull Charset charset) throws IOException {
        URLConnection urlConn = url.openConnection();
        urlConn.setReadTimeout(READ_TIMEOUT);

        return streamToString(urlConn.getInputStream(), charset.toString());
    }

    @Nonnull
    public static byte[] urlToByteArray(@Nonnull URL url) throws IOException {
        URLConnection urlConn = url.openConnection();
        urlConn.setReadTimeout(READ_TIMEOUT);

        return streamToByteArray(urlConn.getInputStream());
    }

    public static void urlToFile(@Nonnull URL url, @Nonnull File file) throws IOException {
        URLConnection urlConn = url.openConnection();
        urlConn.setReadTimeout(READ_TIMEOUT);

        Files.copy(urlConn.getInputStream(), file.toPath());
    }

    /** Also puts the file in a unique temporary directory. */
    @Nonnull
    public static File urlToTmpFile(@Nonnull URL url) throws IOException {
        Path tmpDir = Files.createTempDirectory(null);
        File tmpFile = new File(tmpDir.toFile(), url.getFile());

        Runtime.getRuntime().addShutdownHook(new Thread() {
            public void run() {
                try {
                    Files.delete(tmpFile.toPath());
                    Files.delete(tmpDir);
                } catch (IOException e) {
                    System.err.println("Failed to delete temp files on shutdown. Will be ignored.");
                }
            }
        });

        URLConnection urlConn = url.openConnection();
        urlConn.setReadTimeout(READ_TIMEOUT);

        Files.copy(urlConn.getInputStream(), tmpFile.toPath());

        return tmpFile;
    }

    public static void streamToFile(@Nonnull InputStream is, @Nonnull File outFile) {
        //old, BAD version  (breaks binary files)
//        try {
//            FileWriter fw = new FileWriter(outFile);
//
//            BufferedReader br = new BufferedReader(new InputStreamReader(is));
//            String line = br.readLine();
//            while (line != null) {
//                fw.write(line);
//                line = br.readLine();
//            }
//
//            fw.close();
//        } catch (IOException e) {
//            throw new RuntimeException("Error reading inputstream '" + is + "' or writing to file \"" + outFile.getName() + "\": " + e.getMessage(), e);
//        }
        final Path destination = outFile.toPath();
        try {
            Files.copy(is, destination);
        } catch (IOException e) {
            throw new RuntimeException("Error reading inputstream '" + is + "' or writing to file \"" + outFile.getName() + "\": " + e.getMessage(), e);
        }
    }

    @Nonnull
    public static String currentStackToStacktraceString() {
        Throwable e;
        try {
            throw new RuntimeException();
        } catch (RuntimeException ex) {
            e = ex;
        }
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        e.printStackTrace(pw);
        pw.close();
        String stacktrace = sw.getBuffer().toString();
        return stacktrace;
    }

    @Nonnull
    public static String exceptionToStacktraceString(@Nonnull Throwable e) {
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        e.printStackTrace(pw);
        pw.close();
        String stacktrace = sw.getBuffer().toString();
        return stacktrace;
    }

    /** Exception class + first local classname in stacktrace. */
    @Nonnull
    public static String exceptionToShortDescription(@Nonnull Throwable e) {
        String localClassName = null;
        String methodName = null;
        int lineNr = -1;
        for (final StackTraceElement ste : e.getStackTrace()) {
            if (ste.getClassName() != null && ste.getClassName().contains("be.iminds.ilabt.")) {
                localClassName = ste.getClassName();
                methodName = ste.getMethodName();
                lineNr = ste.getLineNumber();
                break;
            }
        }

        String res = e.getClass().getSimpleName();
        if (localClassName != null) {
            res += " in " + localClassName.replaceAll(".*\\.", "");
            if (methodName != null) {
                res += "." + methodName+"()";
            }
            if (lineNr > 0) {
                res += " @ line " + lineNr;
            }
        } else {
            res += " "+e.getMessage();
        }
        return res;
    }

    @Nonnull
    public static String threadToStacktraceString(@Nonnull Thread thread) {
        StackTraceElement[] stackTraceElements = thread.getStackTrace();
        String res = "";
        for (StackTraceElement e : stackTraceElements) {
            res += e+"\n";
        }
        return res;
    }

    public static void stringToFile(@Nonnull String filename, @Nonnull String text) {
        stringToFile(new File(filename), text);
    }

    public static void stringToFile(@Nonnull File file, @Nonnull String text) {
        try {
            stringToFileWithEx(file, text);
        } catch (IOException e) {
            throw new RuntimeException("Error writing '" + file + "': " + e.getMessage(), e);
        }
    }
    public static void stringToFileWithEx(@Nonnull File file, @Nonnull String text) throws IOException {
        try (FileWriter fw = new FileWriter(file)) {
            fw.write(text);
        }
    }

    /**
     * @returns false if group or others have any permissions, or if userRead/userWrite/userExecute is true but the permission on the file is not
     */
    public static boolean checkUserOnlyPerms(@Nonnull File file,
                                             boolean userRead,
                                             boolean userWrite,
                                             boolean userExecute) throws IOException {
        if (OSDetector.os == OSDetector.OS.UNIX || OSDetector.os == OSDetector.OS.MAC) {
            Set<PosixFilePermission> perms = Files.getPosixFilePermissions(file.toPath());
            if (!checkNoPermsForGroupAndOther(file)) return false;
            if (userRead && !perms.contains(PosixFilePermission.OWNER_READ)) return false;
            if (userWrite && !perms.contains(PosixFilePermission.OWNER_WRITE)) return false;
            if (userExecute && !perms.contains(PosixFilePermission.OWNER_EXECUTE)) return false;
        }
        return true;
    }

    public static boolean checkNoPermsForGroupAndOther(@Nonnull File file) throws IOException {
        if (OSDetector.os == OSDetector.OS.UNIX || OSDetector.os == OSDetector.OS.MAC) {
            Set<PosixFilePermission> perms = Files.getPosixFilePermissions(file.toPath());
            if (perms.contains(PosixFilePermission.GROUP_READ)) return false;
            if (perms.contains(PosixFilePermission.GROUP_WRITE)) return false;
            if (perms.contains(PosixFilePermission.GROUP_EXECUTE)) return false;
            if (perms.contains(PosixFilePermission.OTHERS_READ)) return false;
            if (perms.contains(PosixFilePermission.OTHERS_WRITE)) return false;
            if (perms.contains(PosixFilePermission.OTHERS_EXECUTE)) return false;
        }
        return true;
    }

    /**
     * group or others have NO permissions
     */
    public static void assureUserOnlyPerms(@Nonnull File file, boolean userRead, boolean userWrite, boolean userExecute) throws IOException {
        if (OSDetector.os == OSDetector.OS.UNIX || OSDetector.os == OSDetector.OS.MAC) {
            boolean needToChange = false;
            Set<PosixFilePermission> perms = Files.getPosixFilePermissions(file.toPath());
            if (perms.contains(PosixFilePermission.GROUP_READ)) needToChange = true;
            if (perms.contains(PosixFilePermission.GROUP_WRITE)) needToChange = true;
            if (perms.contains(PosixFilePermission.GROUP_EXECUTE)) needToChange = true;
            if (perms.contains(PosixFilePermission.OTHERS_READ)) needToChange = true;
            if (perms.contains(PosixFilePermission.OTHERS_WRITE)) needToChange = true;
            if (perms.contains(PosixFilePermission.OTHERS_EXECUTE)) needToChange = true;
            if (perms.contains(PosixFilePermission.OWNER_READ) != userRead) needToChange = true;
            if (perms.contains(PosixFilePermission.OWNER_WRITE) != userWrite) needToChange = true;
            if (perms.contains(PosixFilePermission.OWNER_EXECUTE) != userExecute) needToChange = true;
            if (needToChange) setUserOnlyPerms(file, userRead, userWrite, userExecute);
        }
    }

    /**
     * group or others have NO permissions
     */
    public static void setUserOnlyPerms(@Nonnull String file, boolean userRead, boolean userWrite, boolean userExecute) throws IOException {
        setUserOnlyPerms(new File(file), userRead, userWrite, userExecute);
    }

    /**
     * group or others have NO permissions
     */
    public static void setUserOnlyPerms(@Nonnull File file, boolean userRead, boolean userWrite, boolean userExecute) throws IOException {
        if (OSDetector.os == OSDetector.OS.UNIX || OSDetector.os == OSDetector.OS.MAC) {
            //this methods works in Java 7
            Set<PosixFilePermission> perms = new HashSet<>();
            if (userRead) perms.add(PosixFilePermission.OWNER_READ);
            if (userWrite) perms.add(PosixFilePermission.OWNER_WRITE);
            if (userExecute) perms.add(PosixFilePermission.OWNER_EXECUTE);
            Files.setPosixFilePermissions(file.toPath(), perms);
        }
    }

    public static void setAllowAllRead(@Nonnull File file, boolean userWrite, boolean userExecute) throws IOException {
        if (OSDetector.os == OSDetector.OS.UNIX || OSDetector.os == OSDetector.OS.MAC) {
            //this methods works in Java 7
            Set<PosixFilePermission> perms = new HashSet<>();
            perms.add(PosixFilePermission.OWNER_READ);
            perms.add(PosixFilePermission.GROUP_READ);
            perms.add(PosixFilePermission.OTHERS_READ);
            if (userWrite) perms.add(PosixFilePermission.OWNER_WRITE);
            if (userExecute) perms.add(PosixFilePermission.OWNER_EXECUTE);
            Files.setPosixFilePermissions(file.toPath(), perms);
        }
    }

    public static void deleteDirAtExit(File gitRepoDir) {
        Runtime.getRuntime().addShutdownHook(new Thread() {
            public void run() {
                try {
                    Files.walkFileTree(gitRepoDir.toPath(), new SimpleFileVisitor<Path>() {
                        @Override
                        public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                            try {
                                Files.delete(file);
                            } catch (IOException e) {
                                System.err.println("Failed to delete file '"+file+"' in temp dir on shutdown. Will be ignored.");
                                //ignore errors
                            }
                            return FileVisitResult.CONTINUE;
                        }

                        @Override
                        public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
                            try {
                                Files.delete(dir);
                            } catch (IOException e) {
                                System.err.println("Failed to delete dir '"+dir+"' in temp dir on shutdown. Will be ignored.");
                                //ignore errors
                            }
                            return FileVisitResult.CONTINUE;
                        }
                    });
                } catch (IOException e) {
                    System.err.println("Failed to delete temp dir '"+gitRepoDir.getPath()+"' on shutdown. Will be ignored.");
                }
            }
        });
    }


    public static class PathParts {
        @Nonnull public final String dir;
        @Nonnull public final String fileBase;
        @Nonnull public final String ext;

        public PathParts(@Nonnull String dir, @Nonnull String fileBase, @Nonnull String ext) {
            this.dir = dir;
            this.fileBase = fileBase;
            this.ext = ext;
        }

        @Nonnull
        public String getFullExt() {
            if (ext.isEmpty())
                return "";
            else
                return "."+ext;
        }
        @Nonnull
        public String getFullPath() {
            if (ext.isEmpty())
                return dir+fileBase;
            else
                return dir+fileBase+"."+ext;
        }
        @Nonnull
        public String getBasenameAndExt() {
            if (ext.isEmpty())
                return fileBase;
            else
                return fileBase+"."+ext;
        }
    }

    @Nonnull
    public static PathParts splitUnixPath(@Nonnull String remoteUnixPath) {
        int dirSlashIndex = remoteUnixPath.lastIndexOf('/');
        String dir,filename;
        if (dirSlashIndex >= 0) {
            dir = remoteUnixPath.substring(0, dirSlashIndex+1);
            filename = remoteUnixPath.substring(dirSlashIndex + 1);
        } else {
            dir = "";
            filename = remoteUnixPath;
        }

        String fileBase, ext;
        int extDotIndex = filename.lastIndexOf('.');
        if (extDotIndex > 0) {
            fileBase = filename.substring(0, extDotIndex);
            ext = filename.substring(extDotIndex+1);
        } else {
            fileBase = filename;
            ext = "";
        }

        PathParts res = new PathParts(dir, fileBase, ext);
        assert res.getFullPath().equals(remoteUnixPath) : "Could not reconstruct: "+res.getFullPath()+" != "+remoteUnixPath;
        return res;
    }


    /**
     * @param fullPath
     * @return the filename of the file specified in fullPath (removes leading directories, and keeps any extension)
     */
    @Nonnull
    public static String getUnixBasename(@Nonnull String fullPath) {
        if (fullPath.length() > 1 && fullPath.endsWith("/")) {
            //special case: dir that ends with / should result in dir basename, not in empty string
            fullPath = fullPath.substring(0, fullPath.length()-1);
        }
        int separatorLocation = fullPath.lastIndexOf('/');
        return fullPath.substring(separatorLocation + 1);
    }

    /**
     * @param fullPath
     * @return the path of the file specified in fullPath (removes filename part, leaving all directoriess)
     */
    @Nonnull
    public static String getUnixPathOnly(@Nonnull String fullPath) {
        int separatorLocation = fullPath.lastIndexOf('/');
        if (separatorLocation < 0) return fullPath;
        return fullPath.substring(0, separatorLocation+1);
    }
}
